@@ -250,7 +250,7 @@ static int check_fcntl(void)
 		return -1;
 
 	if (fcntl(fd, F_GETOWNER_UIDS, (long)v)) {
-		pr_perror("Can'r fetch file owner UIDs");
+		pr_perror("Can't fetch file owner UIDs");
 		close(fd);
 		return -1;
 	}
@@ -1062,6 +1062,25 @@ static int check_can_map_vdso(void)
 	return -1;
 }
 
+static int check_sk_netns(void)
+{
+	if (kerndat_socket_netns() < 0)
+		return -1;
+
+	if (!kdat.sk_ns)
+		return -1;
+
+	return 0;
+}
+
+static int check_sk_unix_file(void)
+{
+	if (!kdat.sk_unix_file)
+		return -1;
+
+	return 0;
+}
+
 static int (*chk_feature)(void);
 
 /*
@@ -1168,6 +1187,7 @@ int cr_check(void)
 		ret |= check_can_map_vdso();
 		ret |= check_uffd();
 		ret |= check_uffd_noncoop();
+		ret |= check_sk_netns();
 	}
 
 	/*
@@ -1197,6 +1217,39 @@ static int check_tun(void)
 	return check_tun_cr(-1);
 }
 
+static int check_tun_netns(void)
+{
+	bool has = false;
+	check_tun_netns_cr(&has);
+	return has ? 0 : -1;
+}
+
+static int check_nsid(void)
+{
+	if (kerndat_nsid() < 0)
+		return -1;
+
+	if (!kdat.has_nsid) {
+		pr_warn("NSID isn't supported\n");
+		return -1;
+	}
+
+	return 0;
+}
+
+static int check_link_nsid(void)
+{
+	if (kerndat_link_nsid() < 0)
+		return -1;
+
+	if (!kdat.has_link_nsid) {
+		pr_warn("NSID isn't supported\n");
+		return -1;
+	}
+
+	return 0;
+}
+
 struct feature_list {
 	char *name;
 	int (*func)();
@@ -1208,6 +1261,7 @@ static struct feature_list feature_list[] = {
 	{ "aio_remap", check_aio_remap },
 	{ "timerfd", check_timerfd },
 	{ "tun", check_tun },
+	{ "tun_ns", check_tun_netns },
 	{ "userns", check_userns },
 	{ "fdinfo_lock", check_fdinfo_lock },
 	{ "seccomp_suspend", check_ptrace_suspend_seccomp },
@@ -1220,6 +1274,10 @@ static struct feature_list feature_list[] = {
 	{ "uffd", check_uffd },
 	{ "uffd-noncoop", check_uffd_noncoop },
 	{ "can_map_vdso", check_can_map_vdso},
+	{ "sk_ns", check_sk_netns },
+	{ "sk_unix_file", check_sk_unix_file },
+	{ "nsid", check_nsid },
+	{ "link_nsid", check_link_nsid},
 	{ NULL, NULL },
 };
 
