@@ -38,6 +38,7 @@
 #include "page-xfer.h"
 #include "common/lock.h"
 #include "rst-malloc.h"
+#include "fdstore.h"
 #include "util.h"
 
 #undef  LOG_PREFIX
@@ -58,34 +59,37 @@
 
 #define LAZY_PAGES_RESTORE_FINISHED	0x52535446	/* ReSTore Finished */
 
+/*
+ * Backround transfer parameters.
+ * The default xfer length is arbitraty set to 64Kbytes
+ * The limit of 4Mbytes matches the maximal chunk size we can have in
+ * a pipe in the page-server
+ */
+#define DEFAULT_XFER_LEN (64 << 10)
+#define MAX_XFER_LEN (4 << 20)
+
 static mutex_t *lazy_sock_mutex;
 
 struct lazy_iov {
 	struct list_head l;
-	unsigned long base;	/* run-time start address, tracks remaps */
-	unsigned long img_base;	/* start address at the dump time */
-	unsigned long len;
-	bool queued;
-};
-
-struct lp_req {
-	unsigned long addr;	/* actual #PF (or background) destination */
-	unsigned long img_addr;	/* the corresponding address at the dump time */
-	unsigned long len;
-	struct list_head l;
+	unsigned long start;	/* run-time start address, tracks remaps */
+	unsigned long end;	/* run-time end address, tracks remaps */
+	unsigned long img_start;	/* start address at the dump time */
 };
 
 struct lazy_pages_info {
 	int pid;
+	bool exited;
 
 	struct list_head iovs;
 	struct list_head reqs;
 
 	struct lazy_pages_info *parent;
-	unsigned num_children;
+	unsigned ref_cnt;
 
 	struct page_read pr;
 
+	unsigned long xfer_len;	/* in pages */
 	unsigned long total_pages;
 	unsigned long copied_pages;
 
@@ -93,6 +97,7 @@ struct lazy_pages_info {
 
 	struct list_head l;
 
+	unsigned long buf_size;
 	void *buf;
 };
 
@@ -103,6 +108,8 @@ static LIST_HEAD(pending_lpis);
 static int epollfd;
 static bool restore_finished;
 static struct epoll_rfd lazy_sk_rfd;
+/* socket for communication with lazy-pages daemon */
+static int lazy_pages_sk_id = -1;
 
 static int handle_uffd_event(struct epoll_rfd *lpfd);
 
@@ -119,6 +126,8 @@ static struct lazy_pages_info *lpi_init(void)
 	INIT_LIST_HEAD(&lpi->reqs);
 	INIT_LIST_HEAD(&lpi->l);
 	lpi->lpfd.read_event = handle_uffd_event;
+	lpi->xfer_len = DEFAULT_XFER_LEN;
+	lpi->ref_cnt = 1;
 
 	return lpi;
 }
@@ -131,23 +140,43 @@ static void free_iovs(struct lazy_pages_info *lpi)
 		list_del(&p->l);
 		xfree(p);
 	}
+
+	list_for_each_entry_safe(p, n, &lpi->reqs, l) {
+		list_del(&p->l);
+		xfree(p);
+	}
+}
+
+static void lpi_fini(struct lazy_pages_info *lpi);
+
+static inline void lpi_put(struct lazy_pages_info *lpi)
+{
+	lpi->ref_cnt--;
+	if (!lpi->ref_cnt)
+		lpi_fini(lpi);
+}
+
+static inline void lpi_get(struct lazy_pages_info *lpi)
+{
+	lpi->ref_cnt++;
 }
 
 static void lpi_fini(struct lazy_pages_info *lpi)
 {
 	if (!lpi)
 		return;
-	free(lpi->buf);
+	xfree(lpi->buf);
 	free_iovs(lpi);
 	if (lpi->lpfd.fd > 0)
 		close(lpi->lpfd.fd);
 	if (lpi->parent)
-		lpi->parent->num_children--;
-	if (!lpi->parent && !lpi->num_children && lpi->pr.close)
+		lpi_put(lpi->parent);
+	if (!lpi->parent && lpi->pr.close)
 		lpi->pr.close(&lpi->pr);
-	free(lpi);
+	xfree(lpi);
 }
 
+
 static int prepare_sock_addr(struct sockaddr_un *saddr)
 {
 	int len;
@@ -173,7 +202,7 @@ static int send_uffd(int sendfd, int pid)
 	if (sendfd < 0)
 		return -1;
 
-	fd = get_service_fd(LAZY_PAGES_SK_OFF);
+	fd = fdstore_get(lazy_pages_sk_id);
 	if (fd < 0) {
 		pr_err("%s: get_service_fd\n", __func__);
 		return -1;
@@ -292,8 +321,7 @@ err:
 
 int prepare_lazy_pages_socket(void)
 {
-	int fd, new_fd;
-	int len;
+	int fd, len, ret = -1;
 	struct sockaddr_un sun;
 
 	if (!opts.lazy_pages)
@@ -311,19 +339,22 @@ int prepare_lazy_pages_socket(void)
 	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
 		return -1;
 
-	new_fd = install_service_fd(LAZY_PAGES_SK_OFF, fd);
-	close(fd);
-	if (new_fd < 0)
-		return -1;
-
 	len = offsetof(struct sockaddr_un, sun_path) + strlen(sun.sun_path);
-	if (connect(new_fd, (struct sockaddr *) &sun, len) < 0) {
+	if (connect(fd, (struct sockaddr *) &sun, len) < 0) {
 		pr_perror("connect to %s failed", sun.sun_path);
-		close(new_fd);
-		return -1;
+		goto out;
 	}
 
-	return 0;
+	lazy_pages_sk_id = fdstore_add(fd);
+	if (lazy_pages_sk_id < 0) {
+		pr_perror("Can't add fd to fdstore");
+		goto out;
+	}
+
+	ret = 0;
+out:
+	close(fd);
+	return ret;
 }
 
 static int server_listen(struct sockaddr_un *saddr)
@@ -378,13 +409,13 @@ static struct lazy_iov *find_iov(struct lazy_pages_info *lpi,
 	struct lazy_iov *iov;
 
 	list_for_each_entry(iov, &lpi->iovs, l)
-		if (addr >= iov->base && addr < iov->base + iov->len)
+		if (addr >= iov->start && addr < iov->end)
 			return iov;
 
 	return NULL;
 }
 
-static int split_iov(struct lazy_iov *iov, unsigned long addr, bool new_below)
+static int split_iov(struct lazy_iov *iov, unsigned long addr)
 {
 	struct lazy_iov *new;
 
@@ -392,46 +423,84 @@ static int split_iov(struct lazy_iov *iov, unsigned long addr, bool new_below)
 	if (!new)
 		return -1;
 
-	if (new_below) {
-		new->base = iov->base;
-		new->img_base = iov->img_base;
-		new->len = addr - iov->base;
-		iov->base = addr;
-		iov->img_base += new->len;
-		iov->len -= new->len;
-		list_add_tail(&new->l, &iov->l);
-	} else {
-		new->base = addr;
-		new->img_base = iov->img_base + addr - iov->base;
-		new->len = iov->len - (addr - iov->base);
-		iov->len -= new->len;
-		list_add(&new->l, &iov->l);
+	new->start = addr;
+	new->img_start = iov->img_start + addr - iov->start;
+	new->end = iov->end;
+	iov->end = addr;
+	list_add(&new->l, &iov->l);
+
+	return 0;
+}
+
+static void iov_list_insert(struct lazy_iov *new, struct list_head *dst)
+{
+	struct lazy_iov *iov;
+
+	if (list_empty(dst)) {
+		list_move(&new->l, dst);
+		return;
 	}
 
+	list_for_each_entry(iov, dst, l) {
+		if (new->start < iov->start) {
+			list_move_tail(&new->l, &iov->l);
+			break;
+		}
+		if (list_is_last(&iov->l, dst) &&
+		    new->start > iov->start) {
+			list_move(&new->l, &iov->l);
+			break;
+		}
+	}
+}
+
+static void merge_iov_lists(struct list_head *src, struct list_head *dst)
+{
+	struct lazy_iov *iov, *n;
+
+	if (list_empty(src))
+		return;
+
+	list_for_each_entry_safe(iov, n, src, l)
+		iov_list_insert(iov, dst);
+}
+
+static int __copy_iov_list(struct list_head *src, struct list_head *dst)
+{
+	struct lazy_iov *iov, *new;
+
+	list_for_each_entry(iov, src, l) {
+		new = xzalloc(sizeof(*new));
+		if (!new)
+			return -1;
+
+		new->start = iov->start;
+		new->img_start = iov->img_start;
+		new->end = iov->end;
+
+		list_add_tail(&new->l, dst);
+	}
+
+
 	return 0;
 }
 
 static int copy_iovs(struct lazy_pages_info *src, struct lazy_pages_info *dst)
 {
-	struct lazy_iov *iov, *new;
-	int max_iov_len = 0;
+	if (__copy_iov_list(&src->iovs, &dst->iovs))
+		goto free_iovs;
 
-	list_for_each_entry(iov, &src->iovs, l) {
-		new = xzalloc(sizeof(*new));
-		if (!new)
-			return -1;
+	if (__copy_iov_list(&src->reqs, &dst->reqs))
+		goto free_iovs;
 
-		new->base = iov->base;
-		new->img_base = iov->img_base;
-		new->len = iov->len;
+	/*
+	 * The IOVs aready in flight for the parent process need to be
+	 * transferred again for the child process
+	 */
+	merge_iov_lists(&dst->reqs, &dst->iovs);
 
-		list_add_tail(&new->l, &dst->iovs);
-
-		if (new->len > max_iov_len)
-			max_iov_len = new->len;
-	}
-
-	if (posix_memalign(&dst->buf, PAGE_SIZE, max_iov_len))
+	dst->buf_size = src->buf_size;
+	if (posix_memalign(&dst->buf, PAGE_SIZE, dst->buf_size))
 		goto free_iovs;
 
 	return 0;
@@ -445,13 +514,13 @@ free_iovs:
  * Purge range (addr, addr + len) from lazy_iovs. The range may
  * cover several continuous IOVs.
  */
-static int drop_iovs(struct lazy_pages_info *lpi, unsigned long addr, int len)
+static int __drop_iovs(struct list_head *iovs, unsigned long addr, int len)
 {
 	struct lazy_iov *iov, *n;
 
-	list_for_each_entry_safe(iov, n, &lpi->iovs, l) {
-		unsigned long start = iov->base;
-		unsigned long end = start + iov->len;
+	list_for_each_entry_safe(iov, n, iovs, l) {
+		unsigned long start = iov->start;
+		unsigned long end = iov->end;
 
 		if (len <= 0 || addr + len < start)
 			break;
@@ -464,31 +533,28 @@ static int drop_iovs(struct lazy_pages_info *lpi, unsigned long addr, int len)
 			addr = start;
 		}
 
-		iov->queued = false;
-
 		/*
 		 * The range completely fits into the current IOV.
-		 * If addr equals iov_base we just "drop" the
+		 * If addr equals iov_start we just "drop" the
 		 * beginning of the IOV. Otherwise, we make the IOV to
 		 * end at addr, and add a new IOV start starts at
 		 * addr + len.
 		 */
 		if (addr + len < end) {
 			if (addr == start) {
-				iov->base += len;
-				iov->img_base += len;
-				iov->len -= len;
+				iov->start += len;
+				iov->img_start += len;
 			} else {
-				if (split_iov(iov, addr + len, false))
+				if (split_iov(iov, addr + len))
 					return -1;
-				iov->len -= len;
+				iov->end = addr;
 			}
 			break;
 		}
 
 		/*
 		 * The range spawns beyond the end of the current IOV.
-		 * If addr equals iov_base we just "drop" the entire
+		 * If addr equals iov_start we just "drop" the entire
 		 * IOV.  Otherwise, we cut the beginning of the IOV
 		 * and continue to the next one with the updated range
 		 */
@@ -496,7 +562,7 @@ static int drop_iovs(struct lazy_pages_info *lpi, unsigned long addr, int len)
 			list_del(&iov->l);
 			xfree(iov);
 		} else {
-			iov->len -= (end - addr);
+			iov->end = addr;
 		}
 
 		len -= (end - addr);
@@ -506,56 +572,91 @@ static int drop_iovs(struct lazy_pages_info *lpi, unsigned long addr, int len)
 	return 0;
 }
 
-static int remap_iovs(struct lazy_pages_info *lpi, unsigned long from,
-		      unsigned long to, unsigned long len)
+static int drop_iovs(struct lazy_pages_info *lpi, unsigned long addr, int len)
+{
+	if (__drop_iovs(&lpi->iovs, addr, len))
+		return -1;
+
+	if (__drop_iovs(&lpi->reqs, addr, len))
+		return -1;
+
+	return 0;
+}
+
+
+static struct lazy_iov *extract_range(struct lazy_iov *iov,
+				      unsigned long start,
+				      unsigned long end)
+{
+	/* move the IOV tail into a new IOV */
+	if (end < iov->end)
+		if (split_iov(iov, end))
+			return NULL;
+
+	if (start == iov->start)
+		return iov;
+
+	/* after splitting the IOV head we'll need the ->next IOV */
+	if (split_iov(iov, start))
+		return NULL;
+
+	return list_entry(iov->l.next, struct lazy_iov, l);
+}
+
+static int __remap_iovs(struct list_head *iovs,	unsigned long from,
+			unsigned long to, unsigned long len)
 {
-	unsigned long off = to - from;
-	struct lazy_iov *iov, *n, *p;
 	LIST_HEAD(remaps);
 
-	list_for_each_entry_safe(iov, n, &lpi->iovs, l) {
-		unsigned long iov_end = iov->base + iov->len;
+	unsigned long off = to - from;
+	struct lazy_iov *iov, *n;
 
-		if (from >= iov_end)
+	list_for_each_entry_safe(iov, n, iovs, l) {
+		if (from >= iov->end)
 			continue;
 
-		if (len <= 0 || from + len < iov->base)
+		if (len <= 0 || from + len <= iov->start)
 			break;
 
-		if (from < iov->base) {
-			len -= (iov->base - from);
-			from = iov->base;
+		if (from < iov->start) {
+			len -= (iov->start - from);
+			from = iov->start;
 		}
 
-		if (from > iov->base)
-			if (split_iov(iov, from, true))
+		if (from > iov->start) {
+			if (split_iov(iov, from))
 				return -1;
-		if (from + len < iov_end)
-			if (split_iov(iov, from + len, false))
+			list_safe_reset_next(iov, n, l);
+			continue;
+		}
+
+		if (from + len < iov->end) {
+			if (split_iov(iov, from + len))
 				return -1;
+			list_safe_reset_next(iov, n, l);
+		}
 
-		list_safe_reset_next(iov, n, l);
-
-		/* here we have iov->base = from, iov_end <= from + len */
-		from = iov_end;
-		len -= iov->len;
-		iov->base += off;
+		/* here we have iov->start = from, iov->end <= from + len */
+		from = iov->end;
+		len -= iov->end - iov->start;
+		iov->start += off;
+		iov->end += off;
 		list_move_tail(&iov->l, &remaps);
 	}
 
-	list_for_each_entry_safe(iov, n, &remaps, l) {
-		list_for_each_entry(p, &lpi->iovs, l) {
-			if (iov->base < p->base) {
-				list_move_tail(&iov->l, &p->l);
-				break;
-			}
-			if (list_is_last(&p->l, &lpi->iovs) &&
-			    iov->base > p->base) {
-				list_move(&iov->l, &p->l);
-				break;
-			}
-		}
-	}
+	merge_iov_lists(&remaps, iovs);
+
+	return 0;
+}
+
+static int remap_iovs(struct lazy_pages_info *lpi, unsigned long from,
+		      unsigned long to, unsigned long len)
+{
+	if (__remap_iovs(&lpi->iovs, from, to, len))
+		return -1;
+
+	if (__remap_iovs(&lpi->reqs, from, to, len))
+		return -1;
 
 	return 0;
 }
@@ -600,9 +701,9 @@ static int collect_iovs(struct lazy_pages_info *lpi)
 				goto free_iovs;
 
 			len = min_t(uint64_t, end, vma->end) - start;
-			iov->base = start;
-			iov->img_base = start;
-			iov->len = len;
+			iov->start = start;
+			iov->img_start = start;
+			iov->end = iov->start + len;
 			list_add_tail(&iov->l, &lpi->iovs);
 
 			if (len > max_iov_len)
@@ -615,7 +716,8 @@ static int collect_iovs(struct lazy_pages_info *lpi)
 		}
 	}
 
-	if (posix_memalign(&lpi->buf, PAGE_SIZE, max_iov_len))
+	lpi->buf_size = max_iov_len;
+	if (posix_memalign(&lpi->buf, PAGE_SIZE, lpi->buf_size))
 		goto free_iovs;
 
 	ret = nr_pages;
@@ -703,7 +805,8 @@ static int handle_exit(struct lazy_pages_info *lpi)
 		return -1;
 	free_iovs(lpi);
 	close(lpi->lpfd.fd);
-	lpi->lpfd.fd = 0;
+	lpi->lpfd.fd = -lpi->lpfd.fd;
+	lpi->exited = true;
 
 	/* keep it for tracking in-flight requests and for the summary */
 	list_move_tail(&lpi->l, &lpis);
@@ -758,32 +861,54 @@ static int uffd_io_complete(struct page_read *pr, unsigned long img_addr, int nr
 {
 	struct lazy_pages_info *lpi;
 	unsigned long addr = 0;
-	struct lp_req *req;
+	int req_pages, ret;
+	struct lazy_iov *req;
 
 	lpi = container_of(pr, struct lazy_pages_info, pr);
 
-	list_for_each_entry(req, &lpi->reqs, l) {
-		if (req->img_addr == img_addr) {
-			addr = req->addr;
-			list_del(&req->l);
-			xfree(req);
-			break;
-		}
-	}
-
 	/*
 	 * The process may exit while we still have requests in
 	 * flight. We just drop the request and the received data in
 	 * this case to avoid making uffd unhappy
 	 */
-	if (list_empty(&lpi->iovs))
-	    return 0;
+	if (lpi->exited)
+		return 0;
 
-	BUG_ON(!addr);
+	list_for_each_entry(req, &lpi->reqs, l) {
+		if (req->img_start == img_addr) {
+			addr = req->start;
+			break;
+		}
+	}
 
-	if (uffd_copy(lpi, addr, nr))
-		return -1;
+	/* the request may be already gone because if unmap/remove */
+	if (!addr)
+		return 0;
 
+	/*
+	 * By the time we get the pages from the remote source, parts
+	 * of the request may already be gone because of unmap/remove
+	 * OTOH, the remote side may send less pages than we requested.
+	 * Make sure we are not trying to uffd_copy more memory than
+	 * we should.
+	 */
+	req_pages = (req->end - req->start) / PAGE_SIZE;
+	nr = min(nr, req_pages);
+
+	ret = uffd_copy(lpi, addr, nr);
+	if (ret < 0)
+		return ret;
+
+	/* recheck if the process exited, it may be detected in uffd_copy */
+	if (lpi->exited)
+		return 0;
+
+	/*
+	 * Since the completed request length may differ from the
+	 * actual data we've received we re-insert the request to IOVs
+	 * list and let drop_iovs do the range math, free memory etc.
+	 */
+	iov_list_insert(req, &lpi->iovs);
 	return drop_iovs(lpi, addr, nr * PAGE_SIZE);
 }
 
@@ -808,7 +933,7 @@ static int uffd_zero(struct lazy_pages_info *lpi, __u64 address, int nr_pages)
 /*
  * Seek for the requested address in the pagemap. If it is found, the
  * subsequent call to pr->page_read will bring us the data. If the
- * address is not found in the pagemap, but no error occured, the
+ * address is not found in the pagemap, but no error occurred, the
  * address should be mapped to zero pfn.
  *
  * Returns 0 for zero pages, 1 for "real" pages and negative value on
@@ -846,54 +971,51 @@ static int uffd_handle_pages(struct lazy_pages_info *lpi, __u64 address, int nr,
 	return 0;
 }
 
-static struct lazy_iov *first_pending_iov(struct lazy_pages_info *lpi)
+static struct lazy_iov *pick_next_range(struct lazy_pages_info *lpi)
 {
-	struct lazy_iov *iov;
-
-	list_for_each_entry(iov, &lpi->iovs, l)
-		if (!iov->queued)
-			return iov;
-
-	return NULL;
+	return list_first_entry(&lpi->iovs, struct lazy_iov, l);
 }
 
-static bool is_iov_queued(struct lazy_pages_info *lpi, struct lazy_iov *iov)
+/*
+ * This is very simple heurstics for backgroud transfer control.
+ * The idea is to transfer larger chunks when there is no page faults
+ * and drop the backgroud transfer size each time #PF occurs to some
+ * default value. The default is empirically set to 64Kbytes
+ */
+static void update_xfer_len(struct lazy_pages_info *lpi, bool pf)
 {
-	struct lp_req *req;
+	if (pf)
+		lpi->xfer_len = DEFAULT_XFER_LEN;
+	else
+		lpi->xfer_len += DEFAULT_XFER_LEN;
 
-	list_for_each_entry(req, &lpi->reqs, l)
-		if (req->addr >= iov->base && req->addr < iov->base + iov->len)
-			return true;
-
-	return false;
+	if (lpi->xfer_len > MAX_XFER_LEN)
+		lpi->xfer_len = MAX_XFER_LEN;
 }
 
-static int handle_remaining_pages(struct lazy_pages_info *lpi)
+static int xfer_pages(struct lazy_pages_info *lpi)
 {
 	struct lazy_iov *iov;
-	struct lp_req *req;
-	int nr_pages, err;
+	unsigned int nr_pages;
+	unsigned long len;
+	int err;
 
-	iov = first_pending_iov(lpi);
+	iov = pick_next_range(lpi);
 	if (!iov)
 		return 0;
 
-	if (is_iov_queued(lpi, iov))
-		return 0;
+	len = min(iov->end - iov->start, lpi->xfer_len);
 
-	nr_pages = iov->len / PAGE_SIZE;
-
-	req = xzalloc(sizeof(*req));
-	if (!req)
+	iov = extract_range(iov, iov->start, iov->start + len);
+	if (!iov)
 		return -1;
+	list_move(&iov->l, &lpi->reqs);
 
-	req->addr = iov->base;
-	req->img_addr = iov->img_base;
-	req->len = iov->len;
-	list_add(&req->l, &lpi->reqs);
-	iov->queued = true;
+	nr_pages = (iov->end - iov->start) / PAGE_SIZE;
 
-	err = uffd_handle_pages(lpi, req->img_addr, nr_pages, PR_ASYNC | PR_ASAP);
+	update_xfer_len(lpi, false);
+
+	err = uffd_handle_pages(lpi, iov->img_start, nr_pages, PR_ASYNC | PR_ASAP);
 	if (err < 0) {
 		lp_err(lpi, "Error during UFFD copy\n");
 		return -1;
@@ -971,7 +1093,7 @@ static int handle_fork(struct lazy_pages_info *parent_lpi, struct uffd_msg *msg)
 
 	dup_page_read(&lpi->parent->pr, &lpi->pr);
 
-	lpi->parent->num_children++;
+	lpi_get(lpi->parent);
 
 	return 1;
 
@@ -980,10 +1102,18 @@ out:
 	return -1;
 }
 
+/*
+ * We may exit epoll_run_rfds() loop because of non-fork() event. In
+ * such case we return 1 rather than 0 to let the caller know that no
+ * fork() events were pending
+ */
 static int complete_forks(int epollfd, struct epoll_event **events, int *nr_fds)
 {
 	struct lazy_pages_info *lpi, *n;
 
+	if (list_empty(&pending_lpis))
+		return 1;
+
 	list_for_each_entry(lpi, &pending_lpis, l)
 		(*nr_fds)++;
 
@@ -1004,10 +1134,10 @@ static int complete_forks(int epollfd, struct epoll_event **events, int *nr_fds)
 
 static bool is_page_queued(struct lazy_pages_info *lpi, unsigned long addr)
 {
-	struct lp_req *req;
+	struct lazy_iov *req;
 
 	list_for_each_entry(req, &lpi->reqs, l)
-		if (addr >= req->addr && addr < req->addr + req->len)
+		if (addr >= req->start && addr < req->end)
 			return true;
 
 	return false;
@@ -1015,7 +1145,6 @@ static bool is_page_queued(struct lazy_pages_info *lpi, unsigned long addr)
 
 static int handle_page_fault(struct lazy_pages_info *lpi, struct uffd_msg *msg)
 {
-	struct lp_req *req;
 	struct lazy_iov *iov;
 	__u64 address;
 	int ret;
@@ -1031,15 +1160,15 @@ static int handle_page_fault(struct lazy_pages_info *lpi, struct uffd_msg *msg)
 	if (!iov)
 		return uffd_zero(lpi, address, 1);
 
-	req = xzalloc(sizeof(*req));
-	if (!req)
+	iov = extract_range(iov, address, address + PAGE_SIZE);
+	if (!iov)
 		return -1;
-	req->addr = address;
-	req->img_addr = iov->img_base + (address - iov->base);
-	req->len = PAGE_SIZE;
-	list_add(&req->l, &lpi->reqs);
 
-	ret = uffd_handle_pages(lpi, req->img_addr, 1, PR_ASYNC | PR_ASAP);
+	list_move(&iov->l, &lpi->reqs);
+
+	update_xfer_len(lpi, true);
+
+	ret = uffd_handle_pages(lpi, iov->img_start, 1, PR_ASYNC | PR_ASAP);
 	if (ret < 0) {
 		lp_err(lpi, "Error during regular page copy\n");
 		return -1;
@@ -1104,53 +1233,45 @@ static void lazy_pages_summary(struct lazy_pages_info *lpi)
 #endif
 }
 
-#define POLL_TIMEOUT 1000
-
 static int handle_requests(int epollfd, struct epoll_event *events, int nr_fds)
 {
 	struct lazy_pages_info *lpi, *n;
-	/* FIXME -- timeout should decrease over time...  */
-	int poll_timeout = POLL_TIMEOUT;
+	int poll_timeout = -1;
 	int ret;
 
 	for (;;) {
-		bool remaining = false;
-
 		ret = epoll_run_rfds(epollfd, events, nr_fds, poll_timeout);
 		if (ret < 0)
 			goto out;
 		if (ret > 0) {
-			if (complete_forks(epollfd, &events, &nr_fds))
-				return -1;
-			continue;
+			ret = complete_forks(epollfd, &events, &nr_fds);
+			if (ret < 0)
+				goto out;
+			if (restore_finished)
+				poll_timeout = 0;
+			if (!restore_finished || !ret)
+				continue;
 		}
 
-		/* don't start backround fetch before restore is finished */
-		if (!restore_finished)
-			continue;
+		/* make sure we return success if there is nothing to xfer */
+		ret = 0;
 
-		if (poll_timeout)
-			pr_debug("Start handling remaining pages\n");
-
-		poll_timeout = 0;
 		list_for_each_entry_safe(lpi, n, &lpis, l) {
-			if (list_empty(&lpi->iovs) && list_empty(&lpi->reqs)) {
-				lazy_pages_summary(lpi);
-				list_del(&lpi->l);
-				lpi_fini(lpi);
-				continue;
-			}
-
-			remaining = true;
-			if (!list_empty(&lpi->iovs)) {
-				ret = handle_remaining_pages(lpi);
+			if (!list_empty(&lpi->iovs) && list_empty(&lpi->reqs)) {
+				ret = xfer_pages(lpi);
 				if (ret < 0)
 					goto out;
 				break;
 			}
+
+			if (list_empty(&lpi->reqs)) {
+				lazy_pages_summary(lpi);
+				list_del(&lpi->l);
+				lpi_put(lpi);
+			}
 		}
 
-		if (!remaining)
+		if (list_empty(&lpis))
 			break;
 	}
 
@@ -1167,7 +1288,7 @@ int lazy_pages_finish_restore(void)
 	if (!opts.lazy_pages)
 		return 0;
 
-	fd = get_service_fd(LAZY_PAGES_SK_OFF);
+	fd = fdstore_get(lazy_pages_sk_id);
 	if (fd < 0) {
 		pr_err("No lazy-pages socket\n");
 		return -1;
@@ -1213,7 +1334,7 @@ static int lazy_sk_read_event(struct epoll_rfd *rfd)
 		return 0;
 
 	if (ret != sizeof(fin)) {
-		pr_perror("Failed getting restore finished inidication");
+		pr_perror("Failed getting restore finished indication");
 		return -1;
 	}
 
@@ -1224,7 +1345,7 @@ static int lazy_sk_read_event(struct epoll_rfd *rfd)
 
 	restore_finished = true;
 
-	return 0;
+	return 1;
 }
 
 static int lazy_sk_hangup_event(struct epoll_rfd *rfd)
